@@ -285,6 +285,9 @@ sub emit_comparison_template {
     # or even worse within a category inside it. We strip out the
     # extra '../' that is used to generate a relative path which is no
     # longer valid.
+    # Sadly this change is permanent and will influence successive
+    # responses in a permanent execution environment like fcgid or
+    # as a systemd socket service.  Thus we need to revert it later.
     foreach my $cat(@{$key->{'comparecategories'}}) {
         foreach my $service(@{$cat->{'services'}}) {
             foreach my $node(@{$service->{'nodes'}}) {
@@ -292,6 +295,8 @@ sub emit_comparison_template {
                               cimgday cimgweek cimgmonth cimgyear
                               zoomday zoomweek zoommonth zoomyear)) {
                     next unless defined($node->{$imgsrc});
+                    # keep a copy of the original value (to be restored below)
+                    $node->{"orig_$imgsrc"} = $node->{$imgsrc};
                     $node->{$imgsrc} =~ s|^\.\./\.\./(?:\.\./)?|../|;
                 }
             }
@@ -322,13 +327,32 @@ sub emit_comparison_template {
 									NWARNING => scalar(@{$htmlconfig->{"problems"}->{"warnings"}}),
 									NUNKNOWN => scalar(@{$htmlconfig->{"problems"}->{"unknowns"}}),
     );
+
+    # store template output before reverting links
+    my $template_output = $comparisontemplates{$t}->output;
+
+    # restore the paths to their original value
+    foreach my $cat(@{$key->{'comparecategories'}}) {
+        foreach my $service(@{$cat->{'services'}}) {
+            foreach my $node(@{$service->{'nodes'}}) {
+                foreach my $imgsrc(qw(imgday imgweek imgmonth imgyear
+                              cimgday cimgweek cimgmonth cimgyear
+                              zoomday zoomweek zoommonth zoomyear)) {
+                    next unless defined($node->{$imgsrc});
+                    $node->{$imgsrc} = $node->{"orig_$imgsrc"};
+                    delete($node->{"orig_$imgsrc"});
+                }
+            }
+        }
+    }
+
     if($emit_to_stdout){
-		print $comparisontemplates{$t}->output;
+        print $template_output;
 	} else {
 		ensure_dir_exists($file);
 	    open(my $FILE, '>', $file)
     	    or die "Cannot open $file for writing: $!";
-	    print $FILE $comparisontemplates{$t}->output;
+        print $FILE $template_output;
     	close $FILE;
 	}
 }
@@ -402,12 +426,18 @@ sub emit_category_template {
 
     DEBUG "[DEBUG] Creating global category page ".$filename;
 
+    # Manipulate the relative paths for the requested root-level context.
+    # Sadly this change is permanent and will influence successive
+    # responses in a permanent execution environment like fcgid or
+    # as a systemd socket service.  Thus we need to revert it later.
     foreach my $graphs(@{$key->{'graphs'}}) {
         foreach my $graph(@{$graphs->{'graphs'}}) {
             foreach my $imgsrc(qw(imgday imgweek imgmonth imgyear
                               cimgday cimgweek cimgmonth cimgyear
                               zoomday zoomweek zoommonth zoomyear)) {
-                $graph->{$imgsrc} =~ s|^(?:\.\./)+||
+                # keep a copy of the original value (to be restored below)
+                $graph->{"orig_$imgsrc"} = $graph->{$imgsrc};
+                $graph->{$imgsrc} =~ s|^(?:\.\./)+||;
             }
         }
     }
@@ -432,13 +462,28 @@ sub emit_category_template {
 						  NUNKNOWN => scalar(@{$htmlconfig->{"problems"}->{"unknowns"}}),
                          );
 
+    # store template output before reverting links
+    my $template_output = $graphtemplate->output;
+
+    # restore the paths to their original value
+    foreach my $graphs(@{$key->{'graphs'}}) {
+        foreach my $graph(@{$graphs->{'graphs'}}) {
+            foreach my $imgsrc(qw(imgday imgweek imgmonth imgyear
+                              cimgday cimgweek cimgmonth cimgyear
+                              zoomday zoomweek zoommonth zoomyear)) {
+                $graph->{$imgsrc} = $graph->{"orig_$imgsrc"};
+                delete($graph->{"orig_$imgsrc"});
+            }
+        }
+    }
+
     if($emit_to_stdout){
-		print $graphtemplate->output;
+        print $template_output;
 	} else {
 		ensure_dir_exists($filename);
 	    open(my $FILE, '>', $filename)
 			or die "Cannot open $filename for writing: $!";
-    	print $FILE $graphtemplate->output;
+        print $FILE $template_output;
 	    close $FILE;
 	}
 }
@@ -465,6 +510,23 @@ sub emit_problem_template {
 
     INFO "[INFO] Creating problem page ".$filename;
 
+    # Manipulate the relative paths for the requested root-level context.
+    # Sadly this change is permanent and will influence successive
+    # responses in a permanent execution environment like fcgid or
+    # as a systemd socket service.  Thus we need to revert it later.
+    foreach my $problem_graphs (values %{$htmlconfig->{"problems"}}) {
+        foreach my $graph(@$problem_graphs) {
+            foreach my $imgsrc(qw(imgday imgweek imgmonth imgyear
+                                  cimgday cimgweek cimgmonth cimgyear
+                                  zoomday zoomweek zoommonth zoomyear)) {
+                next unless defined($graph->{$imgsrc});
+                # keep a copy of the original value (to be restored below)
+                $graph->{"orig_$imgsrc"} = $graph->{$imgsrc};
+                $graph->{$imgsrc} =~ s|^(?:\.\./)+||;
+            }
+        }
+    }
+
     $graphtemplate->param(
                           CSS_NAME    => get_css_name(),
                           R_PATH      => ".",
@@ -483,13 +545,29 @@ sub emit_problem_template {
 						  NUNKNOWN => scalar(@{$htmlconfig->{"problems"}->{"unknowns"}}),
                          );
 
+    # store template output before reverting links
+    my $template_output = $graphtemplate->output;
+
+    # restore the paths to their original value
+    foreach my $problem_graphs (values %{$htmlconfig->{"problems"}}) {
+        foreach my $graph(@$problem_graphs) {
+            foreach my $imgsrc(qw(imgday imgweek imgmonth imgyear
+                                  cimgday cimgweek cimgmonth cimgyear
+                                  zoomday zoomweek zoommonth zoomyear)) {
+                next unless defined($graph->{$imgsrc});
+                $graph->{$imgsrc} = $graph->{"orig_$imgsrc"};
+                delete($graph->{"orig_$imgsrc"});
+            }
+        }
+    }
+
     if($emit_to_stdout){
-		print $graphtemplate->output;
+        print $template_output;
 	} else {
 		ensure_dir_exists($filename);
 	    open(my $FILE, '>', $filename)
 			or die "Cannot open $filename for writing: $!";
-    	print $FILE $graphtemplate->output;
+        print $FILE $template_output;
 	    close $FILE;
 	}
 }
@@ -925,7 +1003,7 @@ Munin-html creates the html pages.
 
 =head1 VERSION
 
-This is munin-html version 2.0.47
+This is munin-html version 2.0.49
 
 =head1 AUTHORS
 
