@@ -1,7 +1,7 @@
 package Munin::Master::Utils;
 # -*- cperl -*-
 
-# $Id: Utils.pm 3236 2009-12-18 06:25:03Z feiner.tom $
+# $Id$
 
 use strict;
 use warnings;
@@ -16,27 +16,35 @@ use Munin::Common::Defaults;
 use Munin::Master::Logger;
 use Munin::Master::Config;
 use Munin::Common::Config;
-use Log::Log4perl qw (:easy);
+use Log::Log4perl qw(:easy);
 use POSIX qw(strftime);
-use POSIX ":sys_wait_h";
-use RRDs;
-use Symbol 'gensym';
+use POSIX qw(:sys_wait_h);
+use Symbol qw(gensym);
+use Data::Dumper;
+use Storable;
+use Scalar::Util qw(isweak weaken);
 
 our (@ISA, @EXPORT);
 
 @ISA = ('Exporter');
-@EXPORT = ('munin_trend',
-	   'munin_fetch',
+@EXPORT = (
 	   'munin_nscasend',
 	   'munin_createlock',
 	   'munin_removelock',
 	   'munin_runlock',
 	   'munin_getlock',
-	   'munin_readconfig',
+	   'munin_readconfig_raw',
 	   'munin_writeconfig',
+	   'munin_writeconfig_storable',
+	   'munin_read_storable',
+	   'munin_write_storable',
 	   'munin_delete',
 	   'munin_overwrite',
-	   'munin_config',
+	   'munin_dumpconfig',
+	   'munin_dumpconfig_as_str',
+	   'munin_readconfig_base',
+	   'munin_readconfig_part',
+	   'munin_configpart_revision',
 	   'munin_draw_field',
 	   'munin_get_bool',
 	   'munin_get_bool_val',
@@ -46,17 +54,23 @@ our (@ISA, @EXPORT);
 	   'munin_node_status',
 	   'munin_category_status',
 	   'munin_get_picture_filename',
+	   'munin_get_picture_loc',
 	   'munin_get_html_filename',
 	   'munin_get_filename',
+	   'munin_get_keypath',
 	   'munin_graph_column_headers',
 	   'munin_get_max_label_length',
 	   'munin_get_field_order',
 	   'munin_get_rrd_filename',
 	   'munin_get_node_name',
+	   'munin_get_orig_node_name',
 	   'munin_get_parent_name',
+	   'munin_get_node_fqn',
+	   'munin_find_node_by_fqn',
 	   'munin_get_node_loc',
 	   'munin_get_node',
 	   'munin_set_var_loc',
+	   'munin_set_var_path',
 	   'munin_set',
 	   'munin_copy_node_toloc',
 	   'munin_get_separated_node',
@@ -67,6 +81,7 @@ our (@ISA, @EXPORT);
 	   'munin_get_children',
 	   'munin_get_node_partialpath',
 	   'munin_has_subservices',
+	   'munin_get_host_path_from_string',
 	   'print_version_and_exit',
 	   'exit_if_run_by_super_user',
 	   'look_for_child',
@@ -77,6 +92,27 @@ my $VERSION = $Munin::Common::Defaults::MUNIN_VERSION;
 
 my $nsca = new IO::Handle;
 my $config = undef;
+my $config_parts = {
+	# config parts that might be loaded and reloaded at times
+	'datafile' => {
+		'timestamp' => 0,
+		'config' => undef,
+		'revision' => 0,
+		'include_base' => 1,
+	},
+	'limits' => {
+		'timestamp' => 0,
+		'config' => undef,
+		'revision' => 0,
+		'include_base' => 1,
+	},
+	'htmlconf' => {
+		'timestamp' => 0,
+		'config' => undef,
+		'revision' => 0,
+		'include_base' => 0,
+	},
+};
 
 my $configfile="$Munin::Common::Defaults::MUNIN_CONFDIR/munin.conf";
 
@@ -87,26 +123,6 @@ my @dircomponents = split('/',$0);
 my $me = pop(@dircomponents);
 
 
-sub munin_trend {
-    my (@array) = @_;
-    return ($array[$#array] - $array[0]);
-}
-
-
-sub munin_fetch {
-    my ($file,$last,$type) = @_;
-    my ($start,$step,$names,$data) = RRDs::fetch $file,$type || "AVERAGE";
-    unless (defined $data)
-    {
-        WARN ("[WARNING] Could not fetch data from $file(".($type||"AVERAGE")."): ". RRDs::error);
-        return;
-    }
-    my @array = map { @$_[0] } splice(@$data, $#$data - ($last || 1));
-    return $array[0] if (!$last);
-    return @array;
-}
-
-
 sub munin_draw_field {
     my $hash   = shift;
 
@@ -230,11 +246,11 @@ sub munin_overwrite {
 
     my ($configfile,$overwrite) = @_;
     for my $key (keys %$overwrite) {
-        next if $key =~ /^#%#/;
+	next if substr($key,0,3) eq '#%#';
 	if (ref $overwrite->{$key}) {
 	    if (!defined $configfile->{$key}) {
 		if (ref $overwrite->{$key} eq "HASH") {
-		    $configfile->{$key}->{'#%#parent'} = $configfile;
+		    $configfile->{$key}->{'#%#parent'} = $configfile; weaken($configfile->{$key}->{'#%#parent'});
 		    $configfile->{$key}->{'#%#name'}   = $key;
 		    munin_overwrite($configfile->{$key},$overwrite->{$key});
 		} else {
@@ -250,36 +266,27 @@ sub munin_overwrite {
     return ($configfile);
 }
 
+sub munin_readconfig_raw {
+    my ($conf, $missingok) = @_;
 
-sub munin_readconfig {
-    my ($conf, $missingok, $corruptok) = @_;
-
-    my $config = undef;
-    my @contents = undef;
+    my $part = undef;
 
     $conf ||= $configfile;
-    if (! -r $conf and ! $missingok) {
-	WARN "munin_readconfig: cannot open '$conf'";
-	return;
-    }
-    if (open (my $CFG, '<', $conf)) {
-	@contents = <$CFG>;
-	close ($CFG);
-        $config = munin_parse_config (\@contents);
+    # try first to read storable version
+    $part = munin_read_storable("$conf.storable");
+    if (!defined $part) {
+        if (! -r $conf and ! $missingok) {
+            WARN "munin_readconfig: cannot open '$conf'";
+            return;
+        }
+        if (open (my $CFG, '<', $conf)) {
+            my @contents = <$CFG>;
+            close ($CFG);
+            $part = munin_parse_config (\@contents);
+        }
     }
 
-    # Some important defaults before we return...
-    $config->{'dropdownlimit'} ||= $Munin::Common::Defaults::DROPDOWNLIMIT;
-    $config->{'rundir'}        ||= $Munin::Common::Defaults::MUNIN_STATEDIR;
-    $config->{'dbdir'}         ||= $Munin::Common::Defaults::MUNIN_DBDIR;
-    $config->{'logdir'}        ||= $Munin::Common::Defaults::MUNIN_LOGDIR;
-    $config->{'tmpldir'}       ||= "$Munin::Common::Defaults::MUNIN_CONFDIR/templates/";
-    $config->{'htmldir'}       ||= $Munin::Common::Defaults::MUNIN_HTMLDIR;
-    $config->{'spooldir'}      ||= $Munin::Common::Defaults::MUNIN_SSPOOLDIR;
-    $config->{'#%#parent'}     = undef;
-    $config->{'#%#name'}       = "root";
-
-    return ($config);
+    return $part;
 }
 
 
@@ -396,28 +403,28 @@ sub munin_get_var_path
     return;
 }
 
- 
+
 sub munin_find_field {
     # Starting at the (presumably the root) $hash make recursive calls
     # until for example graph_title or value is found, and then
-    # continue recursing and itterating to all are found.
+    # continue recursing and itterating until all are found.
     #
     # Then we return a array of pointers into the $hash
+    #
+    # This function will not use REs as they are inordinately
+    # expensive.  There is a munin_find_field_for_limits that will
+    # match REs instead of whole strings.
+
+    my ($hash, $field, $avoid) = @_;
 
-    my $hash  = shift;
-    my $field = shift;
-    my $avoid = shift;
     my $res = [];
 
-    if (ref ($field) ne "Regexp") {
-	$field = qr/^$field$/;
-    }
-
     if (ref ($hash) eq "HASH") {
 	foreach my $key (keys %{$hash}) {
-	    next if $key =~ /^#%#/;
+	    next if substr($key,0,3) eq '#%#';
 	    last if defined $avoid and $key eq $avoid;
-	    if ($key =~ $field) {
+	    # Always check $key eq $field first here, or we break
+	    if ($key eq $field) {
 		push @$res, $hash;
 	    } elsif (ref ($hash->{$key}) eq "HASH") {
 		push @$res, @{munin_find_field ($hash->{$key}, $field, $avoid)};
@@ -430,9 +437,8 @@ sub munin_find_field {
 
 
 sub munin_find_field_for_limits {
-    my $hash  = shift;
-    my $field = shift;
-    my $avoid = shift;
+    my ($hash, $field, $avoid) = @_;
+
     my $res = [];
 
     if (ref ($field) ne "Regexp") {
@@ -441,7 +447,7 @@ sub munin_find_field_for_limits {
 
     if (ref ($hash) eq "HASH") {
 	foreach my $key (keys %{$hash}) {
-	    next if $key =~ /^#%#/;
+	    next if substr($key,0,3) eq '#%#';
 	    last if defined $avoid and $key eq $avoid;
 	    if (ref ($hash->{$key}) eq "HASH") {
 		push @$res, @{munin_find_field_for_limits ($hash->{$key}, $field, $avoid)};
@@ -462,7 +468,7 @@ sub munin_get_children {
     return if (ref ($hash) ne "HASH");
 
     foreach my $key (keys %{$hash}) {
-	next if $key =~ /^#%#/;
+	next if substr($key,0,3) eq '#%#';
 	if (defined $hash->{$key} and ref ($hash->{$key}) eq "HASH") {
 	    push @$res, $hash->{$key};
 	}
@@ -479,7 +485,7 @@ sub munin_get_separated_node
 
     if (ref ($hash) eq "HASH") {
 	foreach my $key (keys %$hash) {
-	    next if $key =~ /^#%#/;
+	    next if substr($key,0,3) eq '#%#';
 	    if (ref ($hash->{$key}) eq "HASH") {
 		$ret->{$key} = munin_get_separated_node ($hash->{$key});
 	    } else {
@@ -505,19 +511,68 @@ sub munin_get_parent_name
     }
 }
 
+sub munin_get_orig_node_name {
+    my $hash = shift;
+
+    if (ref ($hash) eq "HASH" and defined $hash->{'#%#name'}) {
+		return (defined $hash->{'#%#origname'}) ? $hash->{'#%#origname'} : $hash->{'#%#name'};
+    } else { 
+	return;
+    }
+}
 
 sub munin_get_node_name
 {
     my $hash = shift;
 
     if (ref ($hash) eq "HASH" and defined $hash->{'#%#name'}) {
-	return $hash->{'#%#name'};
+		return $hash->{'#%#name'};
     } else { 
 	return;
     }
 }
 
 
+sub munin_get_node_fqn
+{
+    my $hash = shift;
+
+    if (ref ($hash) eq "HASH") {
+	my $fqn = "";
+	if (defined $hash->{'#%#name'}) {
+		$fqn = $hash->{'#%#name'};
+	}
+ 	if (defined $hash->{'#%#parent'}) {
+		# Recursively prepend the parent, concatenation with /
+		$fqn = munin_get_node_fqn ($hash->{'#%#parent'}) . "/" . $fqn;
+	}
+	return $fqn;
+    } else {
+	return;
+    }
+}
+
+
+sub munin_find_node_by_fqn {
+    # The FQN should be relative to the point in the hash we're handed.
+    my($hash,$fqn) = @_;
+
+    my $here = $hash;
+
+    my @path = split('/',$fqn);
+
+    foreach my $pc (@path) {
+	next if $pc eq 'root';
+	if (exists $here->{$pc}) {
+	    $here = $here->{$pc};
+	} else {
+	    confess "Could not find FQN $fqn!";
+	}
+    }
+    return $here ;
+}
+
+
 sub munin_get_picture_loc {
     my $hash = shift;
     my $res = [];
@@ -527,9 +582,12 @@ sub munin_get_picture_loc {
     }
     if (defined $hash->{'#%#origin'}) {
 	$res = munin_get_picture_loc ($hash->{'#%#origin'});
+    } elsif (defined $hash->{'#%#origparent'}){
+        $res = munin_get_picture_loc ($hash->{'#%#origparent'});
+        push @$res, munin_get_orig_node_name ($hash) if defined $res;
     } elsif (defined $hash->{'#%#parent'}) {
-	$res = munin_get_picture_loc ($hash->{'#%#parent'});
-	push @$res, munin_get_node_name ($hash) if defined $res;
+	    $res = munin_get_picture_loc ($hash->{'#%#parent'});
+	    push @$res, munin_get_orig_node_name ($hash) if defined $res;
     }
     return $res;
 }
@@ -540,16 +598,19 @@ sub munin_get_node_loc {
     my $res = [];
 
     if (ref ($hash) ne "HASH") { # Not a has node
-    	return;
+        return;
     }
     if (defined $hash->{'#%#parent'}) {
-	$res = munin_get_node_loc ($hash->{'#%#parent'});
-	push @$res, munin_get_node_name ($hash) if defined $res;
+        if(defined $hash->{'#%#origparent'}){
+            $res = munin_get_node_loc ($hash->{'#%#origparent'});
+        } else {
+            $res = munin_get_node_loc ($hash->{'#%#parent'});
+        }
+        push @$res, munin_get_orig_node_name ($hash) if defined $res;
     }
     return $res;
 }
 
-
 sub munin_get_parent {
     my $hash = shift;
 
@@ -573,11 +634,13 @@ sub munin_get_node {
     my $loc  = shift;
 
     foreach my $tmpvar (@$loc) {
-	if ($tmpvar !~ /\S/) {
-	    ERROR "[ERROR] munin_get_node: Cannot work on hash node \"$tmpvar\"";
+	if (! exists $hash->{$tmpvar} ) {
+	    # Only complain on a blank key if is no key like that. Usually it
+	    # shouln't, so it avoids a needless regexp in this highly used
+	    # function
+	    ERROR "[ERROR] munin_get_node: Cannot work on hash node '$tmpvar'" if ($tmpvar !~ /\S/);
 	    return undef;
-	}
-	return undef if !exists $hash->{$tmpvar};
+        }	
 	$hash = $hash->{$tmpvar};
     }
     return $hash;
@@ -597,20 +660,31 @@ sub munin_set_var_loc
     my $hash = shift;
     my $loc  = shift;
     my $val  = shift;
-    my @aloc = @$loc;
 
-    my $tmpvar = shift @aloc;
-    $tmpvar = shift @aloc while (defined $tmpvar and $tmpvar =~ /^#%#/);
-    if ($tmpvar !~ /\S/) {
+    my $iloc = 0;
+
+    # XXX -  Dirty breaking recursive function 
+    # --> Using goto is BAD, but enough for now
+START:
+
+    # Find the next normal value (that doesn't begin with #%#)
+    my $tmpvar = $loc->[$iloc++];
+    $tmpvar = $loc->[$iloc++] while (defined $tmpvar and
+				 substr($tmpvar,0,3) eq '#%#');
+
+    if (index($tmpvar, " ") != -1) {
 	ERROR "[ERROR] munin_set_var_loc: Cannot work on hash node \"$tmpvar\"";
 	return;
     }
-    if (@aloc > 0) {
-	if (!defined $hash->{$tmpvar} or !defined $hash->{$tmpvar}->{"#%#name"}) { # Init the new node
-	    $hash->{$tmpvar}->{"#%#parent"} = $hash;
+    if (scalar @$loc > $iloc) {
+	if (!defined $hash->{$tmpvar} or !defined $hash->{$tmpvar}->{"#%#name"}) { 
+            # Init the new node
+	    $hash->{$tmpvar}->{"#%#parent"} = $hash; weaken($hash->{$tmpvar}->{"#%#parent"});
 	    $hash->{$tmpvar}->{"#%#name"} = $tmpvar;
 	}
-        return munin_set_var_loc ($hash->{$tmpvar}, \@aloc, $val);
+	# Recurse
+        $hash = $hash->{$tmpvar};
+	goto START;
     } else {
         WARN "[WARNING] munin_set_var_loc: Setting unknown option '$tmpvar' at "
 	    . munin_get_keypath($hash)
@@ -642,9 +716,9 @@ sub munin_get_node_partialpath
 	my @leftarr = split (/;/, $leftstring);
 	my @rightarr = split (/\./, $rightstring);
 	push @$varloc, @leftarr, @rightarr
-    } elsif ($var =~ /^\s*([^;:\.]+)\s*$/) {
+    } elsif ($var =~ /^\s*([^;:.]+)\s*$/) {
 	push @$varloc, $var;
-    } elsif ($var =~ /^\s*(.+)\.([^\.:;]+)$/) {
+    } elsif ($var =~ /^\s*([^.]+)\.([^:;]+)$/) {
 	my ($leftstring, $rightstring) = ($1, $2);
 
 	my @leftarr = split (/;/, $leftstring);
@@ -727,7 +801,7 @@ sub munin_writeconfig_loop {
     my ($hash,$fh,$pre) = @_;
 
     foreach my $key (keys %$hash) {
-	next if $key =~ /#%#/;
+	next if substr($key,0,3) eq '#%#';
 	my $path = (defined $pre ? join(';', ($pre, $key)) : $key);
 	if (ref ($hash->{$key}) eq "HASH") {
 	    munin_writeconfig_loop ($hash->{$key}, $fh, $path);
@@ -747,35 +821,159 @@ sub munin_writeconfig_loop {
     }
 }
 
+sub munin_read_storable {
+	my ($storable_filename, $default) = @_;
+
+	if (-e $storable_filename) {
+		my $storable = eval { Storable::retrieve($storable_filename); };
+		return $storable unless $@;
+
+		# Didn't managed to read storable. 
+		# Removing it as it is already torched anyway.
+		unlink($storable_filename);
+	}
+	
+	# return default if no better option
+	return $default;
+}
+
+sub munin_write_storable {
+	my ($storable_filename, $data) = @_;
+	DEBUG "[DEBUG] about to write '$storable_filename'";
+
+	# We don't need to write anything if there is nothing to write.
+	return unless defined $data;
+
+	my $storable_filename_tmp = $storable_filename . ".tmp.$$";
+
+	# Write datafile.storable, in network order to be architecture indep
+        Storable::nstore($data, $storable_filename_tmp);
+
+	# Atomic commit
+	rename ($storable_filename_tmp, $storable_filename);
+}
+
+sub munin_writeconfig_storable {
+	my ($datafilename,$data) = @_;
+
+	DEBUG "[DEBUG] Writing state to $datafilename";
+
+	munin_write_storable($datafilename, $data);
+}
 
 sub munin_writeconfig {
     my ($datafilename,$data,$fh) = @_;
 
     DEBUG "[DEBUG] Writing state to $datafilename";
 
-    if (!defined $fh) {
+    my $is_fh_already_managed = defined $fh;
+    if (! $is_fh_already_managed) {
 	$fh = gensym();
 	unless (open ($fh, ">", $datafilename)) {
 	    LOGCROAK "Fatal error: Could not open \"$datafilename\" for writing: $!";
 	}
     }
-
+    
     # Write version
     print $fh "version $VERSION\n";
     # Write datafile
     munin_writeconfig_loop ($data, $fh);
+    
+    if (! $is_fh_already_managed) {
+        DEBUG "[DEBUG] Closing filehandle \"$datafilename\"...\n";
+        close ($fh);
+    }
+}
 
-    DEBUG "[DEBUG] Closing filehandle \"$datafilename\"...\n";
-    close ($fh);
+sub munin_dumpconfig_as_str {
+    my ($config) = @_;
+
+    local $Data::Dumper::Sortkeys = sub { [ sort grep {!/^#%#parent$/} keys %{$_[0]} ]; };
+    local $Data::Dumper::Indent = 1;
+
+    return Dumper $config;
 }
 
 
-sub munin_config {
+sub munin_dumpconfig {
+    my ($config) = @_;
+    my $indent = $Data::Dumper::Indent;
+    my $sorter = $Data::Dumper::Sortkeys;
+
+    $Data::Dumper::Sortkeys =
+      sub { [ sort grep {!/^#%#parent$/} keys %{$_[0]} ]; };
+    $Data::Dumper::Indent = 1;
+
+    print "##\n";
+    print "## NOTE: #%#parent is hidden to make the print readable!\n";
+    print "##\n";
+    print Dumper $config;
+
+    $Data::Dumper::Sortkeys = $sorter;
+    $Data::Dumper::Indent = $indent;
+}
+
+sub munin_configpart_revision {
+    my $what = shift;
+    if (defined $what and defined $config_parts->{$what}) {
+	return $config_parts->{$what}{revision};
+    }
+    my $rev = 0;
+    foreach $what (keys %$config_parts) {
+	$rev += $config_parts->{$what}{revision};
+    }
+    return $rev;
+}
+
+sub munin_readconfig_part {
+    my $what = shift;
+    my $missingok = shift;
+    if (! defined $config_parts->{$what}) {
+	ERROR "[ERROR] munin_readconfig_part with unknown part name ($what).";
+	return undef;
+    }
+    # for now, we only really care about storable.
+    # No reason to bother reading non-storable elements anyway.
+    my $filename = "$config->{dbdir}/$what.storable";
+    my $part = {};
+    my $doupdate = 0;
+    if (! -f $filename) {
+	unless (defined $missingok and $missingok) {
+		ERROR "[FATAL] munin_readconfig_part($what) - missing file";
+		exit(1);
+	}
+	# missing ok, return last value if we have one, copy config if not
+	unless (defined $config_parts->{$what}{config}) {
+		# well, not if we shouldn't include the config
+		if ($config_parts->{$what}{include_base}) {
+			$doupdate = 1;
+		}
+	}
+    } else {
+    	my @stat = stat($filename);
+	if ($config_parts->{$what}{timestamp} < $stat[9]) {
+	    # could use _raw if we wanted to read non-storable fallback
+	    $config_parts->{$what}{config} = undef; # Unalloc RAM for old config, since it will be overriden anyway.
+	    $part = munin_read_storable($filename);
+	    $config_parts->{$what}{timestamp} = $stat[9];
+	    $doupdate = 1;
+	}
+    }
+    if ($doupdate) {
+	$part->{'#%#name'} = 'root';
+	$part->{'#%#parent'} = undef;
+	$part = munin_overwrite($part, $config) if ($config_parts->{$what}{include_base});
+	$config_parts->{$what}{config} = $part;
+	++$config_parts->{$what}{revision};
+    }
+    return $config_parts->{$what}{config};
+}
+
+sub munin_readconfig_base {
     my $conffile = shift;
-    my $config   = shift;
 
     $conffile ||= $configfile;
-    $config     = munin_readconfig ($conffile);
+    $config = munin_readconfig_raw($conffile);
 
     if (defined $config->{'includedir'}) {
 	my $dirname = $config->{'includedir'};
@@ -792,15 +990,25 @@ sub munin_config {
 	foreach my $f (@files) {
 	    INFO "Reading additional config from $f";
 
-	    my $extra = munin_readconfig ($f);
-	    $config = munin_overwrite($extra,$config);
+	    my $extra = munin_readconfig_raw ($f);
+	    # let the new values overwrite what we already have
+	    $config = munin_overwrite($config, $extra);
 	}
     }
 
-    my $data    = munin_readconfig("$config->{dbdir}/datafile", 1, 1);
+    # Some important defaults before we return...
+    $config->{'dropdownlimit'} ||= $Munin::Common::Defaults::DROPDOWNLIMIT;
+    $config->{'rundir'}        ||= $Munin::Common::Defaults::MUNIN_STATEDIR;
+    $config->{'dbdir'}         ||= $Munin::Common::Defaults::MUNIN_DBDIR;
+    $config->{'logdir'}        ||= $Munin::Common::Defaults::MUNIN_LOGDIR;
+    $config->{'tmpldir'}       ||= "$Munin::Common::Defaults::MUNIN_CONFDIR/templates/";
+    $config->{'staticdir'}     ||= "$Munin::Common::Defaults::MUNIN_CONFDIR/static/";
+    $config->{'htmldir'}       ||= $Munin::Common::Defaults::MUNIN_HTMLDIR;
+    $config->{'spooldir'}      ||= $Munin::Common::Defaults::MUNIN_SPOOLDIR;
+    $config->{'#%#parent'}     = undef;
+    $config->{'#%#name'}       = "root";
 
-    $data = munin_overwrite($data,$config);
-    return ($data);
+    return $config;
 }
 
 
@@ -824,13 +1032,13 @@ sub munin_get_html_filename {
     } @$loc;
 	
     if (defined $hash->{'graph_title'} and !munin_has_subservices ($hash)) {
-	$plugin = pop @$loc or return;
+        $plugin = pop @$loc or return;
     }
-
+    
     if (@$loc) { # The rest is used as directory names...
-	$ret .= "/" . join ('/', @$loc);
+        $ret .= "/" . join ('/', @$loc);
     }
-
+    
     return "$ret/$plugin.html";
 }
 
@@ -928,6 +1136,7 @@ sub munin_get_keypath {
 	    # In service land, working towards host.
 	    # If i or my parent has a graph_title we're still working with services
 	    if (defined $i->{'#%#parent'}{graph_title} or defined $i->{graph_title}) {
+		$name =~ s/-/_/g; # can't handle dashes in service or below
 		unshift(@service,$name);
 	    } else {
 		$host = 1;
@@ -939,7 +1148,7 @@ sub munin_get_keypath {
     }
 
     if ($asfile) {
-	return join('/',@group).'-'.join('-',@service);
+	return (shift @group).'/'.join('/',@group).'-'.join('-',@service);
     } else {
 	return join(';',@group).':'.join('.',@service);
     }
@@ -1038,14 +1247,20 @@ sub munin_get_bool_val
 
 sub munin_get
 {
-    my $hash   = shift;
-    my $field  = shift;
-    my $default = shift;
+    my ($hash, $field, $default) = @_;
 
-    return $default if (ref ($hash) ne "HASH");
-    return $hash->{$field} if defined $hash->{$field} and ref($hash->{$field}) ne "HASH";
-    return $default if not defined $hash->{'#%#parent'};
-    return munin_get ($hash->{'#%#parent'}, $field, $default);
+    # Iterate to top if needed
+    do {
+        return $default if (ref ($hash) ne "HASH");
+
+        my $hash_field = $hash->{$field};
+        return $hash_field if (defined $hash_field && ref($hash_field) ne "HASH");
+
+        # Go up
+        $hash = $hash->{'#%#parent'};
+    } while (defined $hash);
+
+    return $default;
 }
 
 
@@ -1059,7 +1274,7 @@ sub munin_copy_node_toloc
 
     if (ref ($from) eq "HASH") {
 	foreach my $key (keys %$from) {
-	    next if $key =~ /^#%#/;
+	    next if substr($key,0,3) eq '#%#';
 	    if (ref ($from->{$key}) eq "HASH") {
 		munin_copy_node_toloc ($from->{$key}, $to, [@$loc, $key]);
 	    } else {
@@ -1265,7 +1480,7 @@ sub munin_get_field_order
     for my $fieldnode (@{munin_find_field ($hash, "label")}) {
         my $fieldname = munin_get_node_name ($fieldnode);
 	push @$result,$fieldname
-	    if !grep m[^\Q$fieldname\E(?:=|$)], @$result;;
+	    if !grep m[^\Q$fieldname\E(?:=|$)], @$result;
     }
 
     for my $fieldnode (@{munin_find_field ($hash, "stack")}) {
@@ -1298,7 +1513,6 @@ sub munin_get_rrd_filename {
     my $path    = shift;
 
     my $result  = undef;
-    my $name    = munin_get_node_name ($field);
 
     # Bail out on bad input data
     return if !defined $field or ref ($field) ne "HASH";
@@ -1331,7 +1545,7 @@ sub munin_mkdir_p {
     my ($dirname, $umask) = @_;
 
     eval {
-        mkpath($dirname, {mode => $umask});
+        mkpath($dirname, 0, $umask);
     };
     return if $EVAL_ERROR;
     return 1;
@@ -1413,6 +1627,21 @@ sub munin_has_subservices {
     return $hash->{'#%#has_subservices'};
 }
 
+sub munin_get_host_path_from_string {
+	# splits a host definition, as used in the config, into a group array and a host name
+	my $key = shift;
+	my (@groups) = split(';', $key);
+	my $host = pop(@groups);
+	if(scalar @groups > 0){
+	} else {
+		my @groups = split('.', $key);
+		if(scalar @groups > 1){
+			@groups = ($groups[0]);
+		}
+	}
+	return (\@groups, $host);
+}
+
 1;
 
 __END__
@@ -1444,9 +1673,16 @@ Returns:
  - Failure: undef
 
 
-=item B<munin_config>
+=item B<munin_readconfig_base>
 
+Read configuration file, include dir files, and initialize important
+default values that are optional.
 
+Parameters:
+ - $file: munin.conf filename. If omitted, default filename is used.
+
+Returns:
+ - Success: The $config hash (also cached in module)
 
 =item B<munin_copy_node>
 
@@ -1494,9 +1730,6 @@ Returns:
  - Success: Boolean; true if field will be graphed, false if not
  - Failure: undef
 
-=item B<munin_fetch>
-
-
 
 =item B<munin_field_status>
 
@@ -1609,7 +1842,7 @@ Returns:
 
 =item B<munin_get_max_label_length>
 
-Get the length of the longest labe in a graph.
+Get the length of the longest label in a graph.
 
 Parameters:
  - $hash: the graph in question
@@ -1709,7 +1942,7 @@ Returns:
 =item B<munin_get_picture_loc>
 
 Get location array for hash node for picture purposes. Differs from
-munin_get_node_loc in that it honors #%#origin metadata 
+munin_get_node_loc in that it honors #%#origin metadata
 
 Parameters: 
  - $hash: A ref to the node 
@@ -1789,7 +2022,7 @@ Returns:
 
  munin_mkdir_p('/a/path/', oct('777'));
 
-Make a directory and recursivly any nonexistent directory in the path
+Make a directory and recursively any nonexistent directory in the path
 to it.
 
 
@@ -1822,9 +2055,15 @@ Returns:
  - Failure: undef
 
 
-=item B<munin_readconfig>
+=item B<munin_readconfig_part>
 
+Read a partial configuration
 
+Parameters:
+ - $what: name of the part that should be loaded (datafile or limits)
+
+Returns:
+ - Success: a $config with the $specified part, but overwritten by $config
 
 =item B<munin_removelock>
 
@@ -1890,10 +2129,6 @@ Returns:
  - Failure: The $hash we were handed
 
 
-=item B<munin_trend>
-
-
-
 =item B<munin_writeconfig>
 
 
